@@ -3,6 +3,7 @@ and kivy scatter
 """
 
 import math
+import copy
 
 import matplotlib
 matplotlib.use('Agg')
@@ -14,7 +15,8 @@ from kivy.properties import ObjectProperty, ListProperty, BooleanProperty, Bound
 from kivy.uix.widget import Widget
 from kivy.vector import Vector
 from matplotlib.backends.backend_agg import FigureCanvasAgg
-from matplotlib.transforms import Bbox
+from matplotlib import cbook
+from weakref import WeakKeyDictionary
 from kivy.metrics import dp
 import numpy as np
 
@@ -43,7 +45,16 @@ class MatplotFigureScatter(Widget):
     pos_y_rect_hor=NumericProperty(0)
     pos_x_rect_ver=NumericProperty(0)
     pos_y_rect_ver=NumericProperty(0)   
-
+    invert_rect_ver = BooleanProperty(False)
+    invert_rect_hor = BooleanProperty(False)
+    legend_instance = ObjectProperty(None)
+    legend_do_scroll_x = BooleanProperty(True)
+    legend_do_scroll_y = BooleanProperty(True)    
+    do_pan_x = BooleanProperty(True)
+    do_pan_y = BooleanProperty(True)    
+    do_zoom_x = BooleanProperty(True)
+    do_zoom_y = BooleanProperty(True)  
+    
     def on_figure(self, obj, value):
         self.figcanvas = _FigureCanvas(self.figure, self)
         self.figcanvas._isDrawn = False
@@ -53,6 +64,22 @@ class MatplotFigureScatter(Widget):
         self.width = w
         self.height = h
 
+        if self.figure.axes[0]:
+            #add copy patch
+            ax=self.figure.axes[0]
+            patch_cpy=copy.copy(ax.patch)
+            patch_cpy.set_visible(False)
+            ax.spines[:].set_zorder(10)
+            patch_cpy.set_zorder(9)
+            self.background_patch_copy= ax.add_patch(patch_cpy)
+
+            #set xmin axes attribute
+            self.axes = self.figure.axes[0]
+            
+            #set default xmin/xmax and ymin/ymax
+            self.xmin,self.xmax = self.axes.get_xlim()
+            self.ymin,self.ymax = self.axes.get_ylim()
+            
         # Texture
         self._img_texture = Texture.create(size=(w, h))
 
@@ -73,7 +100,7 @@ class MatplotFigureScatter(Widget):
         self.draw_left_spline=False #available only when fast_draw is True
         self.touch_mode='pan'
         self.hover_on = False
-        self.xsorted = True #to manage x sorted data (if numpy is used)
+        self.xsorted = False #to manage x sorted data
         self.minzoom = dp(40) #minimum pixel distance to apply zoom
         self.multi_xdata=False
         self.multi_xdata_res = dp(20)
@@ -91,6 +118,18 @@ class MatplotFigureScatter(Widget):
         self.lines=[]
         self.scatters=[]
         self.scatter_label = None
+
+        #background 
+        self.background=None
+        self.background_patch_copy=None   
+
+        #manage adjust x and y
+        self.anchor_x = None
+        self.anchor_y = None 
+
+        #manage back and next event
+        self._nav_stack = cbook.Stack()
+        self.set_history_buttons()   
         
         self.bind(size=self._onSize)
 
@@ -111,12 +150,10 @@ class MatplotFigureScatter(Widget):
         #register lines
         self.lines=lines
                 
-        #white background for blit method (fast draw)
-        props = dict(boxstyle='square',edgecolor='w', facecolor='w', alpha=1.0)
-        self.text_background=self.axes.text(0.5, 1.01, 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX', color='w', transform=self.axes.transAxes, bbox=props)       
-        
         #cursor text
-        self.text = self.axes.text(0.52, 1.01, '', transform=self.axes.transAxes, bbox=props)
+        self.text = self.axes.text(1.0, 1.01, '', 
+                                      transform=self.axes.transAxes,
+                                      ha='right')
 
     def register_scatters(self,scatters:list) -> None:
         """ register scatters method
@@ -329,27 +366,21 @@ class MatplotFigureScatter(Widget):
                     self.text.set_text(f"x={x}, y={y}")
 
                 #blit method (always use because same visual effect as draw)
-                self.axes.draw_artist(self.axes.patch)
-                self.axes.draw_artist(self.text_background)
+                if self.background is None:
+                    self.set_cross_hair_visible(False)
+                    self.axes.figure.canvas.draw_idle()
+                    self.axes.figure.canvas.flush_events()                   
+                    self.background = self.axes.figure.canvas.copy_from_bbox(self.axes.figure.bbox)
+                    self.set_cross_hair_visible(True)  
+
+                self.axes.figure.canvas.restore_region(self.background)
                 self.axes.draw_artist(self.text)
-                self.axes.draw_artist(list(self.axes.spines.values())[0])
 
-                for line in self.axes.lines:
-                    self.axes.draw_artist(line)
-                # for scatter in self.axes.scatter:
-                if scatter:
-                    self.axes.draw_artist(scatter)
-                    
-                #find annotation
-                for child in ax.get_children():
-                    if isinstance(child, matplotlib.text.Annotation):               
-                        self.axes.draw_artist(child)
-                    
-
-                mybbox=self.my_blit_box(ax.bbox.bounds[0],ax.bbox.bounds[1],ax.bbox.bounds[2],ax.bbox.bounds[3]+50)                    
+                self.axes.draw_artist(self.horizontal_line)
+                self.axes.draw_artist(self.vertical_line)
                 
                 #draw (blit method)
-                self.axes.figure.canvas.blit(mybbox)                 
+                self.axes.figure.canvas.blit(self.axes.bbox)                
                 self.axes.figure.canvas.flush_events()
 
             #if touch is too far, hide cross hair cursor
@@ -362,12 +393,78 @@ class MatplotFigureScatter(Widget):
         Return:
             None
         """
-        ax = self.axes
-        ax.set_xlim(self.xmin, self.xmax)  
-        ax.set_ylim(self.ymin, self.ymax)                                
+        #do nothing is all min/max are not set
+        if self.xmin is not None and \
+            self.xmax is not None and \
+            self.ymin is not None and \
+            self.ymax is not None:
+                
+            ax = self.axes
+            ax.set_xlim(self.xmin, self.xmax)  
+            ax.set_ylim(self.ymin, self.ymax)                                
+    
+            ax.figure.canvas.draw_idle()
+            ax.figure.canvas.flush_events() 
 
-        ax.figure.canvas.draw_idle()
-        ax.figure.canvas.flush_events() 
+    def back(self, *args):
+        """
+        Move back up the view lim stack.
+        For convenience of being directly connected as a GUI callback, which
+        often get passed additional parameters, this method accepts arbitrary
+        parameters, but does not use them.
+        """
+        self._nav_stack.back()
+        self.set_history_buttons()
+        self._update_view()
+
+    def forward(self, *args):
+        """
+        Move forward in the view lim stack.
+        For convenience of being directly connected as a GUI callback, which
+        often get passed additional parameters, this method accepts arbitrary
+        parameters, but does not use them.
+        """
+        self._nav_stack.forward()
+        self.set_history_buttons()
+        self._update_view()
+
+    def push_current(self):
+       """Push the current view limits and position onto the stack."""
+       self._nav_stack.push(
+           WeakKeyDictionary(
+               {ax: (ax._get_view(),
+                     # Store both the original and modified positions.
+                     (ax.get_position(True).frozen(),
+                      ax.get_position().frozen()))
+                for ax in self.figure.axes}))
+       self.set_history_buttons()       
+
+    def update(self):
+        """Reset the Axes stack."""
+        self._nav_stack.clear()
+        self.set_history_buttons()
+
+    def _update_view(self):
+        """
+        Update the viewlim and position from the view and position stack for
+        each Axes.
+        """
+        nav_info = self._nav_stack()
+        if nav_info is None:
+            return
+        # Retrieve all items at once to avoid any risk of GC deleting an Axes
+        # while in the middle of the loop below.
+        items = list(nav_info.items())
+        for ax, (view, (pos_orig, pos_active)) in items:
+            ax._set_view(view)
+            # Restore both the original and modified positions
+            ax._set_position(pos_orig, 'original')
+            ax._set_position(pos_active, 'active')
+        self.figure.canvas.draw_idle() 
+        self.figure.canvas.flush_events()
+
+    def set_history_buttons(self):
+        """Enable or disable the back/forward button."""
 
     def reset_touch(self) -> None:
         """ reset touch
@@ -377,11 +474,6 @@ class MatplotFigureScatter(Widget):
         """
         self._touches = []
         self._last_touch_pos = {}
-
-    @staticmethod
-    def my_blit_box(x0, y0, width, height):
-        """ build custom matplotlib bbox """
-        return Bbox.from_bounds(x0, y0, width, height)
         
     def _get_scale(self):
         """ kivy scatter _get_scale method """
@@ -432,9 +524,23 @@ class MatplotFigureScatter(Widget):
 
         if len(self._touches) == self.translation_touches:
             if self.touch_mode=='pan':
+                if self._nav_stack() is None:
+                    self.push_current()                
                 self.apply_pan(self.axes, event)
+
+            if self.touch_mode=='pan_x' or self.touch_mode=='pan_y' \
+                or self.touch_mode=='adjust_x' or self.touch_mode=='adjust_y':
+                if self._nav_stack() is None:
+                    self.push_current()                    
+                self.apply_pan(self.axes, event, mode=self.touch_mode)  
+                
+            elif self.touch_mode=='drag_legend':
+                if self.legend_instance:
+                    self.apply_drag_legend(self.axes, event)                
             
             elif self.touch_mode=='zoombox':
+                if self._nav_stack() is None:
+                    self.push_current()                
                 real_x, real_y = event.x - self.pos[0], event.y - self.pos[1]
                 self.draw_box(event, self.x_init,self.y_init, event.x, real_y)
                 
@@ -492,6 +598,20 @@ class MatplotFigureScatter(Widget):
         x, y = event.x, event.y
 
         if self.collide_point(x, y):
+            if self.legend_instance:
+                if self.legend_instance.box.collide_point(x, y):
+                    if self.touch_mode!='drag_legend':
+                        return False   
+                    else:
+                        event.grab(self)
+                        self._touches.append(event)
+                        self._last_touch_pos[event] = event.pos
+                        if len(self._touches)>1:
+                            #new touch, reset background
+                            self.background=None
+                            
+                        return True             
+            
             if event.is_mouse_scrolling:
                 ax = self.axes
                 ax = self.axes
@@ -522,7 +642,10 @@ class MatplotFigureScatter(Widget):
                 event.grab(self)
                 self._touches.append(event)
                 self._last_touch_pos[event] = event.pos
-    
+                if len(self._touches)>1:
+                    #new touch, reset background
+                    self.background=None
+                    
                 return True
 
         else:
@@ -562,7 +685,11 @@ class MatplotFigureScatter(Widget):
             event.ungrab(self)
             del self._last_touch_pos[event]
             self._touches.remove(event)
-
+            if self.touch_mode=='pan' or self.touch_mode=='zoombox' or \
+                self.touch_mode=='pan_x' or self.touch_mode=='pan_y' \
+                or self.touch_mode=='adjust_x' or self.touch_mode=='adjust_y':   
+                self.push_current()
+                
         x, y = event.x, event.y
         if abs(self._box_size[0]) > 1 or abs(self._box_size[1]) > 1 or self.touch_mode=='zoombox':
             self.reset_box()  
@@ -580,7 +707,11 @@ class MatplotFigureScatter(Widget):
             if self.do_update:
                 self.update_lim()            
 
+            self.anchor_x=None
+            self.anchor_y=None
+            
             ax=self.axes
+            self.background=None
             ax.figure.canvas.draw_idle()
             ax.figure.canvas.flush_events()                           
             
@@ -604,19 +735,23 @@ class MatplotFigureScatter(Widget):
         relx = (cur_xlim[1] - xdata) / (cur_xlim[1] - cur_xlim[0])
         rely = (cur_ylim[1] - ydata) / (cur_ylim[1] - cur_ylim[0])
 
-        ax.set_xlim([xdata - new_width * (1 - relx), xdata + new_width * (relx)])
-        ax.set_ylim([ydata - new_height * (1 - rely), ydata + new_height * (rely)])
+        if self.do_zoom_x:
+            ax.set_xlim([xdata - new_width * (1 - relx), xdata + new_width * (relx)])
+        if self.do_zoom_y:
+            ax.set_ylim([ydata - new_height * (1 - rely), ydata + new_height * (rely)])
 
         if self.fast_draw:   
             #use blit method            
-            ax.draw_artist(ax.patch)
-            
-            #if you want the left spline during on_move (slower) 
-            if self.draw_left_spline and not self.scatters:
-                ax.draw_artist(list(ax.spines.values())[0])
-            elif self.scatters:
-                for line in ax.lines:
-                    ax.draw_artist(line)
+            if self.background is None:
+                self.background_patch_copy.set_visible(True)
+                ax.figure.canvas.draw_idle()
+                ax.figure.canvas.flush_events()                   
+                self.background = ax.figure.canvas.copy_from_bbox(ax.figure.bbox)
+                self.background_patch_copy.set_visible(False)  
+            ax.figure.canvas.restore_region(self.background)
+           
+            for line in ax.lines:
+                ax.draw_artist(line)
                 
             for scatter in self.scatters:
                 self.axes.draw_artist(scatter)
@@ -633,7 +768,7 @@ class MatplotFigureScatter(Widget):
             ax.figure.canvas.draw_idle()
             ax.figure.canvas.flush_events()           
 
-    def apply_pan(self, ax, event):
+    def apply_pan(self, ax, event, mode='pan'):
         """ pan method """
         
         trans = ax.transData.inverted()
@@ -644,21 +779,57 @@ class MatplotFigureScatter(Widget):
 
         cur_xlim = ax.get_xlim()
         cur_ylim = ax.get_ylim()
-        cur_xlim -= dx/2
-        cur_ylim -= dy/2
-        ax.set_xlim(cur_xlim)
-        ax.set_ylim(cur_ylim)
+        if not mode=='pan_y' and not mode=='adjust_y':             
+            if mode=='adjust_x':
+                if self.anchor_x is None:
+                    midpoint= (cur_xlim[1] + cur_xlim[0])/2
+                    if xdata>midpoint:
+                        self.anchor_x='left'
+                    else:
+                        self.anchor_x='right'
+                if self.anchor_x=='left':                
+                    if xdata> cur_xlim[0]:
+                        cur_xlim -= dx/2
+                        ax.set_xlim(None,cur_xlim[1])
+                else:
+                    if xdata< cur_xlim[1]:
+                        cur_xlim -= dx/2
+                        ax.set_xlim(cur_xlim[0],None)
+            else:
+                cur_xlim -= dx/2
+                ax.set_xlim(cur_xlim)
+
+        if not mode=='pan_x' and not mode=='adjust_x':
+            if mode=='adjust_y':
+                if self.anchor_y is None:
+                    midpoint= (cur_ylim[1] + cur_ylim[0])/2
+                    if ydata>midpoint:
+                        self.anchor_y='top'
+                    else:
+                        self.anchor_y='bottom'               
+
+                if self.anchor_y=='top':
+                    if ydata> cur_ylim[0]:
+                        cur_ylim -= dy/2   
+                        ax.set_ylim(None,cur_ylim[1])
+                else:
+                    if ydata< cur_ylim[1]:
+                        cur_ylim -= dy/2   
+                        ax.set_ylim(cur_ylim[0],None)                   
+            else:            
+                cur_ylim -= dy/2
+                ax.set_ylim(cur_ylim)
 
         if self.fast_draw: 
             #use blit method
-            ax.draw_artist(ax.patch)
-
-            if self.draw_left_spline and not self.scatters:
-                ax.draw_artist(list(ax.spines.values())[0])
-            elif self.scatters:                
-                for spine in list(ax.spines.values()):
-                    ax.draw_artist(spine)
-
+            if self.background is None:
+                self.background_patch_copy.set_visible(True)
+                ax.figure.canvas.draw_idle()
+                ax.figure.canvas.flush_events()                   
+                self.background = ax.figure.canvas.copy_from_bbox(ax.figure.bbox)
+                self.background_patch_copy.set_visible(False)  
+            ax.figure.canvas.restore_region(self.background)                
+           
             for line in ax.lines:
                 ax.draw_artist(line)
                 
@@ -678,6 +849,42 @@ class MatplotFigureScatter(Widget):
             ax.figure.canvas.draw_idle()
             ax.figure.canvas.flush_events()
 
+    def apply_drag_legend(self, ax, event):
+        """ drag legend method """
+                        
+        dx = event.x - self._last_touch_pos[event][0]
+        if not self.legend_do_scroll_x:
+            dx=0
+        dy = event.y - self._last_touch_pos[event][1]      
+        if not self.legend_do_scroll_y:
+            dy=0        
+        legend = ax.get_legend()
+        if legend is not None:
+        
+            bbox = legend.get_window_extent()
+            legend_x = bbox.xmin
+            legend_y = bbox.ymin
+               
+            loc_in_canvas = legend_x +dx/2, legend_y+dy/2
+            loc_in_norm_axes = legend.parent.transAxes.inverted().transform_point(loc_in_canvas)
+            legend._loc = tuple(loc_in_norm_axes)
+            
+            #use blit method               
+            if self.background is None:
+                ax.get_legend().set_visible(False)
+                ax.figure.canvas.draw_idle()
+                ax.figure.canvas.flush_events()                   
+                self.background = ax.figure.canvas.copy_from_bbox(ax.figure.bbox)
+                ax.get_legend().set_visible(True)
+            ax.figure.canvas.restore_region(self.background)   
+    
+            ax.draw_artist(legend)
+                
+            ax.figure.canvas.blit(ax.bbox)
+            ax.figure.canvas.flush_events() 
+
+            self.legend_instance.update_size()
+
     def zoom_factory(self, event, ax, base_scale=1.1):
         """ zoom with scrolling mouse method """
 
@@ -708,8 +915,10 @@ class MatplotFigureScatter(Widget):
         relx = (cur_xlim[1] - xdata) / (cur_xlim[1] - cur_xlim[0])
         rely = (cur_ylim[1] - ydata) / (cur_ylim[1] - cur_ylim[0])
 
-        ax.set_xlim([xdata - new_width * (1 - relx), xdata + new_width * (relx)])
-        ax.set_ylim([ydata - new_height * (1 - rely), ydata + new_height * (rely)])
+        if self.do_zoom_x:
+            ax.set_xlim([xdata - new_width * (1 - relx), xdata + new_width * (relx)])
+        if self.do_zoom_y:
+            ax.set_ylim([ydata - new_height * (1 - rely), ydata + new_height * (rely)])
 
         ax.figure.canvas.draw_idle()
         ax.figure.canvas.flush_events()    
@@ -730,7 +939,9 @@ class MatplotFigureScatter(Widget):
         hinch = self._height / dpival
         self.figure.set_size_inches(winch, hinch)
         self.figcanvas.resize_event()
-        self.figcanvas.draw()     
+        self.figcanvas.draw()
+        if self.legend_instance:
+            self.legend_instance.update_size()        
 
     def update_lim(self):
         """ update axis lim if zoombox is used"""
@@ -847,6 +1058,15 @@ class MatplotFigureScatter(Widget):
         else:
             self._alpha_hor=0   
             self._alpha_ver=0
+
+        if x1>x0:
+            self.invert_rect_ver=False
+        else:
+            self.invert_rect_ver=True
+        if y1>y0:
+            self.invert_rect_hor=False
+        else:
+            self.invert_rect_hor=True
             
         self._box_pos = x0, y0
         self._box_size = x1 - x0, y1 - y0
@@ -885,10 +1105,10 @@ class _FigureCanvas(FigureCanvasAgg):
 
 from kivy.factory import Factory
 
-Factory.register('MatplotFigure', MatplotFigure)
+Factory.register('MatplotFigure', MatplotFigureScatter)
 
 Builder.load_string('''
-<MatplotFigure>
+<MatplotFigureScatter>
     canvas:
         Color:
             rgba: (1, 1, 1, 1)
@@ -902,8 +1122,12 @@ Builder.load_string('''
             source: 'border.png'
             pos: self._box_pos
             size: self._box_size
-            border: 3, 3, 3, 3
-            
+            border:
+                dp(1) if root.invert_rect_hor else -dp(1), \
+                dp(1) if root.invert_rect_ver else -dp(1), \
+                dp(1) if root.invert_rect_hor else -dp(1), \
+                dp(1) if root.invert_rect_ver else -dp(1)
+                
     canvas.after:            
         #horizontal rectangle left
 		Color:
@@ -911,7 +1135,8 @@ Builder.load_string('''
 		Line:
 			width: dp(1)
 			rectangle:
-				(self.pos_x_rect_hor-dp(3), self.pos_y_rect_hor-dp(20), dp(4),dp(40))            
+				(self.pos_x_rect_hor+dp(1) if root.invert_rect_ver \
+                 else self.pos_x_rect_hor-dp(4),self.pos_y_rect_hor-dp(20), dp(4),dp(40))            
 
         #horizontal rectangle right
 		Color:
@@ -919,7 +1144,8 @@ Builder.load_string('''
 		Line:
 			width: dp(1)
 			rectangle:
-				(self.pos_x_rect_hor-dp(1)+self._box_size[0], self.pos_y_rect_hor-dp(20), dp(4),dp(40))             
+				(self.pos_x_rect_hor-dp(4)+self._box_size[0] if root.invert_rect_ver \
+                 else self.pos_x_rect_hor+dp(1)+self._box_size[0], self.pos_y_rect_hor-dp(20), dp(4),dp(40))             
 
         #vertical rectangle bottom
 		Color:
@@ -927,7 +1153,8 @@ Builder.load_string('''
 		Line:
 			width: dp(1)
 			rectangle:
-				(self.pos_x_rect_ver-dp(20), self.pos_y_rect_ver-dp(1), dp(40),dp(4))            
+				(self.pos_x_rect_ver-dp(20),self.pos_y_rect_ver+dp(1) if root.invert_rect_hor else \
+                 self.pos_y_rect_ver-dp(4), dp(40),dp(4))            
 
         #vertical rectangle top
 		Color:
@@ -935,6 +1162,7 @@ Builder.load_string('''
 		Line:
 			width: dp(1)
 			rectangle:
-				(self.pos_x_rect_ver-dp(20), self.pos_y_rect_ver-dp(3)+self._box_size[1], dp(40),dp(4))             
-
+				(self.pos_x_rect_ver-dp(20),self.pos_y_rect_ver-dp(4)+self._box_size[1] \
+                 if root.invert_rect_hor else self.pos_y_rect_ver+dp(1)+self._box_size[1], \
+                 dp(40),dp(4))
         ''')
